// Re-usable definitions for basic core elements.
// This is designed to be used by doing `incude <library.scad>`

// Set a reasonable default, mostly so support comes out ok.
$fn = 16;

// Thickness of connector wall in axis of connection.
connector_depth = 0.8;
// Gap to add between connector/tab surfaces and cube surfaces.
// This leaves some extra space for printing issues and keeps friction weak
// enough that connectors will go on without needing silly levels of force.
connector_tolerance = 0.1;
// Arbitrary small value used to expand geometry to make sure it will overlap.
// This is useful to avoid glitchy boolean behaviour caused by surfaces meeting
// one another too closely.
overlap = 1;
// Size of sticky-out tab on connector clips in the axis of connection.
slot_clip_length = 2;
// Size of sticky-out tab on connector perpendicular to axis of connection.
slot_clip_width = 0.8;
// Amount of extra cube outside of the v-slot that the tabs clip into.
// This makes the v-slot a bit more robust.
slot_lip = 0.8;
// Amount to trim off of top and bottom lips of v-slots on x and y axis (see
// explanation attached to `vslots_*()`).
slot_lip_fudge = 0.2;
// Size of a basic 1x1x1 cube.
unit = 16;

// Size of outside of connector.
connector_height = unit-(slot_lip+slot_clip_length-slot_clip_width)*2;
// Distance between connected 1x1x1 cubes including connector.
grid = unit+connector_depth+connector_tolerance*2;
// Gap between connected 1x1x1 cubes.
grid_spacing = connector_depth+connector_tolerance*2;
// Distance from bottom to top of plinth.
plinth_height = unit/4;
// Width/height of face left between v-slots.
slot_separation = unit-slot_clip_length*2-slot_clip_width*2;

// Thin 1x1 part that decoration can be attached to the top surface of.
// Positioned as if part of a 1x1x1 cube centred around the origin.
module plinth() difference() {
	// translate([-unit/2, -unit/2, -unit/2]) cube([unit, unit, plinth_height]);
	// vslots_z_negative_bevelled();
	intersection_for (angle = [0, 90]) rotate(angle) {
		rotate(90, [1, 0, 0]) linear_extrude(slot_separation+overlap*2, center=true) polygon([
			[0, plinth_height+slot_separation/4-unit/2],
			[slot_separation/2, plinth_height-unit/2],
			[slot_separation/2, -unit/2],
			[-slot_separation/2, -unit/2],
			[-slot_separation/2, plinth_height-unit/2],
		]);
	}
	vslots_z_negative();
}

// Modules that create the negative shape of v-slots for each side of a 1x1x1
// cube with its centre at the origin.
// Note that slots for sides are slightly different to slots for the top and
// bottom; they include a fudge factor to shave off the top and bottom layer
// (roughly) to make sure support material and thin walls don't mess up
// important dimensions.
module vslots_x_negative() rotate(-90) vslots_y_negative();
module vslots_x_positive() rotate(90) vslots_y_negative();
module vslots_y_negative() union() {
	vslot_general(false);
	mirror([1, 0, 0]) vslot_general(false);
	rotate(90, [0, 1, 0]) vslot_general(true);
	mirror([0, 0, 1]) rotate(90, [0, 1, 0]) vslot_general(true);
}
module vslots_y_positive() mirror([0, 1, 0]) vslots_y_negative();
module vslots_z_negative() rotate(90, [1, 0, 0]) {
	vslot_general(false);
	mirror([1, 0, 0]) vslot_general(false);
	rotate(90, [0, 1, 0]) vslot_general(false);
	mirror([0, 0, 1]) rotate(90, [0, 1, 0]) vslot_general(false);
}
module vslots_z_positive() mirror([0, 0, 1]) vslots_z_negative();

// Similar to `vslots_z_negative()`, except all edges are bevelled (see
// `bevel()`).
module vslots_z_negative_bevelled() {
	vslots_z_negative();
	rotate(90, [1, 0, 0]) for (angle = [0, 90, 180, 270]) rotate(angle, [0, 1, 0]) bevel();
}

// Creates the basic shape of a half v-slot cutout.
// The slot is positioned and rotated to be cut out of the x-negative edge of
// the y-negative face of a cube centred at the origin.
//
// `trim_supports` should be `true` for the top and bottom slots of faces that
// point along the x or y axis.
// When set, an extra layer (roughly) will be trimmed off the top and bottom of
// the face (see explanation attached to `vslots_*()`).
module vslot_general(trim_supports) intersection() {
	fudge_factor = trim_supports ? slot_lip_fudge : 0;

	// Clockwise from top.
	// Cube to cut into has it's corner at the origin and extends along positive
	// x and y.
	linear_extrude(unit+overlap*2, center=true) translate([-unit/2, -unit/2]) polygon([
		// Cutout for clip to sit in.
		[slot_lip+slot_clip_length, slot_lip+slot_clip_length],
		[slot_lip+slot_clip_length*1.5, slot_lip+slot_clip_length/2],
		// Lip clip pulls against.
		// Fudge factor shaves a tiny bit off.
		[slot_lip+slot_clip_length+fudge_factor, slot_lip+fudge_factor],
		[slot_lip+slot_clip_length+fudge_factor, slot_lip/2-fudge_factor],
		// Little ramp on outside to help get connector over lip.
		[slot_lip+slot_clip_length+slot_lip/2, 0],
		// Overlap outside of face to avoid boolean misbehaving.
		[slot_lip+slot_clip_length+slot_lip/2, -overlap],
		[-overlap, -overlap],
		[-overlap, 0],
		[0, 0],
	]);
	
	// Bevel corners so we don't cut beyond this one face in grids of v-slots.
	// rotate(90, [0, 1, 0]) linear_extrude(unit+overlap*2, center=true) polygon([
	// 	[0, 0],
	// 	[unit/2+overlap*2, -unit/2-overlap*2],
	// 	[-unit/2-overlap*2, -unit/2-overlap*2],
	// ]);
}

// Creates the shape of a bevel cutout that removes the overhanging incomplete
// part of a v-slot when only one side of an object has had slots added.
// This is intended for parts where only one surface has v-slots.
module bevel() {
	translate([-unit/2-overlap/2/sin(45), slot_lip+slot_clip_length-overlap/sin(45)/2-unit/2, 0]) linear_extrude(unit+overlap*2, center=true) {
		rotate(45) square((slot_lip+slot_clip_length)/sin(45)+overlap*2, center=true);
	}
}

// Support for bottom lips of v-slots for a rectangle surrounding a given
// number of cubes.
// Positioned for shapes with the x-negative y-negative corner cube center at
// the origin.
module support(units_x, units_y) {
	assert(units_x > 0);
	assert(units_y > 0);

	square_size_inset = [grid*(units_x-1)+unit-2-slot_lip*2, grid*(units_y-1)+unit-2-slot_lip*2];

	translate([1+slot_lip-unit/2, 1+slot_lip-unit/2, -unit/2]) linear_extrude(slot_lip+slot_lip/4+slot_clip_length-0.2) difference() {
		offset(r=1+0.4) square(square_size_inset);
		offset(r=1-0.4) square(square_size_inset);
	}
}

// Create shape of connector centred at the origin.
module connector() union() {
	// Clips in z axis because we can print them cleanly on this axis.
	intersection() {
		// Basic shape of clips.
		linear_extrude(connector_height, center=true) {
			for (x = [0, 1]) mirror([x, 0]) for (y = [0, 1]) mirror([0, y]) {
				// Single clip with centre of connector at origin.
				// Polygon points are relative to the corner of a cube.
				// Clockwise from top.
				translate([-unit/2, connector_tolerance+connector_depth/2]) polygon([
					[slot_lip+slot_clip_length, slot_lip+slot_clip_length-connector_tolerance/sin(45)],
					[slot_lip+slot_clip_length*1.5-connector_tolerance/sin(45), slot_lip+slot_clip_length/2],
					[slot_lip+slot_clip_length-connector_tolerance*sin(45), slot_lip+connector_tolerance*sin(45)],
					[slot_lip+slot_clip_length-connector_tolerance*sin(45), -connector_tolerance],
					[unit/2+overlap, -connector_tolerance],
					[unit/2+overlap, -connector_tolerance-connector_depth],
					[slot_lip+slot_clip_length-slot_clip_width-connector_tolerance*sin(45), -connector_tolerance-connector_depth],
					[slot_lip+slot_clip_length-slot_clip_width-connector_tolerance*sin(45), slot_lip+slot_clip_length-slot_clip_width-connector_tolerance*2],
				]);
			}
		}
		// Cut off corners so that connectors don't interfere.
		safe_volume_side = (unit/2+connector_tolerance+connector_depth/2)/sin(45)-connector_tolerance*2;
		rotate(45, [1, 0, 0]) cube([unit+overlap*2, safe_volume_side, safe_volume_side], center=true);
	}

	// Rails in x axis because we can't print clips properly for this axis, but we
	// still be the connector to keep itself from slipping.
	intersection() {
		// Basic shape of rails.
		rotate(90, [0, 1, 0]) linear_extrude(unit-(slot_lip+slot_clip_length+0.4)*2, center=true) {
			for (x = [0, 1]) mirror([x, 0]) for (y = [0, 1]) mirror([0, y]) {
				// Single rail with centre of connector at origin.
				// Polygon points are relative to the corner of a cube.
				// Clockwise from top.
				translate([-unit/2, connector_tolerance+connector_depth/2]) polygon([
					[slot_lip+slot_clip_length-slot_clip_width, slot_lip/2+slot_clip_width-connector_tolerance],
					[slot_lip+slot_clip_length+slot_lip/2-connector_tolerance*sin(45), -connector_tolerance],
					[slot_lip+slot_clip_length+slot_lip/2-connector_tolerance*sin(45), -connector_tolerance-connector_depth],
					[slot_lip+slot_clip_length-slot_clip_width, -connector_tolerance-connector_depth],
				]);
			}
		}
		// Cut off corners so they don't join to clips and prevent them flexing.
		rotate(45) cube([(unit/2+slot_clip_width-slot_lip-slot_clip_length)*sin(45)*2, (unit/2+slot_clip_width-slot_lip-slot_clip_length)*sin(45)*2, unit*2], center=true);
	}
}
